import phylogenie.typings as pgt
from phylogenie.models import Distribution, StrictBaseModel
from phylogenie.treesimulator import MutationTargetType

Integer = str | int
Scalar = str | pgt.Scalar
ManyScalars = str | pgt.Many[Scalar]
OneOrManyScalars = Scalar | pgt.Many[Scalar]
OneOrMany2DScalars = Scalar | pgt.Many2D[Scalar]


class SkylineParameterModel(StrictBaseModel):
    value: ManyScalars
    change_times: ManyScalars


class SkylineVectorModel(StrictBaseModel):
    value: str | pgt.Many[OneOrManyScalars]
    change_times: ManyScalars


class SkylineMatrixModel(StrictBaseModel):
    value: str | pgt.Many[OneOrMany2DScalars]
    change_times: ManyScalars


SkylineParameter = Scalar | SkylineParameterModel
SkylineVector = str | pgt.Scalar | pgt.Many[SkylineParameter] | SkylineVectorModel
SkylineMatrix = str | pgt.Scalar | pgt.Many[SkylineVector] | SkylineMatrixModel | None


class Event(StrictBaseModel):
    states: str | list[str] | None = None
    rate: SkylineParameter


class Mutation(Event):
    rate_scalers: dict[MutationTargetType, Distribution]



from typing import Callable

from phylogenie.msa import MSA, Sequence
from phylogenie.tree import Tree


def _parse_newick(newick: str) -> Tree:
    newick = newick.strip()
    stack: list[list[Tree]] = []
    current_children: list[Tree] = []
    current_nodes: list[Tree] = []
    i = 0
    while i < len(newick):

        def _parse_chars(stoppers: list[str]) -> str:
            nonlocal i
            chars = ""
            while newick[i] not in stoppers:
                chars += newick[i]
                i += 1
            return chars

        if newick[i] == "(":
            stack.append(current_nodes)
            current_nodes = []
        else:
            id = _parse_chars([":", ",", ")", ";", "["])
            branch_length = None
            if newick[i] == ":":
                i += 1
                branch_length = float(_parse_chars([",", ")", ";", "["]))

            current_node = Tree(id, branch_length)
            for node in current_children:
                current_node.add_child(node)
                current_children = []
            current_nodes.append(current_node)

            if newick[i] == "[":
                i += 1
                features = _parse_chars(["]"]).split(":")
                i += 1
                if features[0] != "&&NHX":
                    raise ValueError(f"Expected '&&NHX' for node features.")
                for feature in features[1:]:
                    key, value = feature.split("=", 1)
                    try:
                        current_node.set(key, eval(value))
                    except Exception as e:
                        raise ValueError(
                            f"Error setting node feature `{key}` to `{value}`: {e}"
                        )

            if newick[i] == ")":
                current_children = current_nodes
                current_nodes = stack.pop()
            elif newick[i] == ";":
                return current_node

        i += 1

    raise ValueError("Newick string is invalid.")


def load_newick(filepath: str) -> Tree | list[Tree]:
    with open(filepath, "r") as file:
        trees = [_parse_newick(newick) for newick in file]
    return trees[0] if len(trees) == 1 else trees


def _to_newick(tree: Tree) -> str:
    children_newick = ",".join([_to_newick(child) for child in tree.children])
    newick = tree.name
    if children_newick:
        newick = f"({children_newick}){newick}"
    if tree.branch_length is not None:
        newick += f":{tree.branch_length}"
    if tree.features:
        reprs = {k: repr(v).replace("'", '"') for k, v in tree.features.items()}
        for k, r in reprs.items():
            if ":" in r or "=" in r or "]" in r:
                raise ValueError(
                    f"Cannot serialize feature `{k}` with value `{r}`: contains reserved characters."
                )
        features = [f"{k}={repr}" for k, repr in reprs.items()]
        newick += f"[&&NHX:{':'.join(features)}]"
    return newick


def dump_newick(trees: Tree | list[Tree], filepath: str) -> None:
    if isinstance(trees, Tree):
        trees = [trees]
    with open(filepath, "w") as file:
        for t in trees:
            file.write(_to_newick(t) + ";\n")


def load_fasta(
    fasta_file: str, extract_time_from_id: Callable[[str], float] | None = None
) -> MSA:
    sequences: list[Sequence] = []
    with open(fasta_file, "r") as f:
        for line in f:
            if not line.startswith(">"):
                raise ValueError(f"Invalid FASTA format: expected '>', got '{line[0]}'")
            id = line[1:].strip()
            time = None
            if extract_time_from_id is not None:
                time = extract_time_from_id(id)
            elif "|" in id:
                try:
                    time = float(id.split("|")[-1])
                except ValueError:
                    pass
            chars = next(f).strip()
            sequences.append(Sequence(id, chars, time))
    return MSA(sequences)



import re
from copy import deepcopy
from enum import Enum
from typing import Type

from numpy.random import Generator

from phylogenie.models import Distribution
from phylogenie.skyline import SkylineParameterLike
from phylogenie.treesimulator.events.contact_tracing import (
    BirthWithContactTracing,
    SamplingWithContactTracing,
)
from phylogenie.treesimulator.events.core import (
    Birth,
    Death,
    Event,
    Migration,
    Sampling,
)
from phylogenie.treesimulator.model import Model

MUTATION_PREFIX = "MUT-"
MUTATIONS_KEY = "MUTATIONS"


def _get_mutation(state: str) -> str | None:
    return state.split(".")[0] if state.startswith(MUTATION_PREFIX) else None


def _get_mutated_state(mutation_id: int, state: str) -> str:
    if state.startswith(MUTATION_PREFIX):
        _, state = state.split(".")
    return f"{MUTATION_PREFIX}{mutation_id}.{state}"


def get_mutation_id(node_name: str) -> int:
    match = re.search(rf"{MUTATION_PREFIX}(\d+)\.", node_name)
    if match:
        return int(match.group(1))
    return 0


class TargetType(str, Enum):
    BIRTH = "birth"
    DEATH = "death"
    MIGRATION = "migration"
    SAMPLING = "sampling"
    MUTATION = "mutation"


class Mutation(Event):
    def __init__(
        self,
        state: str,
        rate: SkylineParameterLike,
        rate_scalers: dict[TargetType, Distribution],
    ):
        super().__init__(state, rate)
        self.rate_scalers = rate_scalers

    def apply(self, model: Model, time: float, rng: Generator) -> None:
        if MUTATIONS_KEY not in model.context:
            model.context[MUTATIONS_KEY] = 0
        model.context[MUTATIONS_KEY] += 1
        mutation_id = model.context[MUTATIONS_KEY]

        individual = self.draw_individual(model, rng)
        model.migrate(individual, _get_mutated_state(mutation_id, self.state), time)

        rate_scalers = {
            target_type: getattr(rng, rate_scaler.type)(**rate_scaler.args)
            for target_type, rate_scaler in self.rate_scalers.items()
        }

        for event in [
            deepcopy(e)
            for e in model.events
            if _get_mutation(self.state) == _get_mutation(e.state)
        ]:
            event.state = _get_mutated_state(mutation_id, event.state)
            if isinstance(event, Birth | BirthWithContactTracing):
                event.child_state = _get_mutated_state(mutation_id, event.child_state)
            elif isinstance(event, Migration):
                event.target_state = _get_mutated_state(mutation_id, event.target_state)
            elif not isinstance(
                event, Mutation | Death | Sampling | SamplingWithContactTracing
            ):
                raise ValueError(
                    f"Mutation not defined for event of type {type(event)}."
                )

            for target_type, rate_scaler in rate_scalers.items():
                if target_type not in TARGETS:
                    raise ValueError(
                        f"Unsupported target type {target_type} for mutation."
                    )
                if isinstance(event, TARGETS[target_type]):
                    event.rate *= rate_scaler

            model.add_event(event)

    def __repr__(self) -> str:
        return f"Mutation(state={self.state}, rate={self.rate})"


TARGETS: dict[TargetType, tuple[Type[Event], ...]] = {
    TargetType.BIRTH: (Birth, BirthWithContactTracing),
    TargetType.DEATH: (Death,),
    TargetType.MIGRATION: (Migration,),
    TargetType.SAMPLING: (Sampling, SamplingWithContactTracing),
    TargetType.MUTATION: (Mutation,),
}



import os
import time
from collections.abc import Iterable, Sequence

import joblib
import numpy as np
from numpy.random import default_rng
from tqdm import tqdm

from phylogenie.io import dump_newick
from phylogenie.tree import Tree
from phylogenie.treesimulator.features import Feature, set_features
from phylogenie.treesimulator.model import Event, Model


def simulate_tree(
    events: Sequence[Event],
    min_tips: int = 1,
    max_tips: int | None = None,
    max_time: float = np.inf,
    init_state: str | None = None,
    sampling_probability_at_present: float = 0.0,
    seed: int | None = None,
    timeout: float = np.inf,
) -> Tree:
    if max_time == np.inf and max_tips is None:
        raise ValueError("Either max_time or max_tips must be specified.")

    if max_time == np.inf and sampling_probability_at_present:
        raise ValueError(
            "sampling_probability_at_present cannot be set when max_time is infinite."
        )

    states = {e.state for e in events if e.state}
    if init_state is None and len(states) > 1:
        raise ValueError(
            "Init state must be provided for models with more than one state."
        )
    elif init_state is None:
        (init_state,) = states
    elif init_state not in states:
        raise ValueError(f"Init state {init_state} not found in event states: {states}")

    rng = default_rng(seed)
    start_clock = time.perf_counter()
    while True:
        model = Model(init_state, events)
        current_time = 0.0
        change_times = sorted(set(t for e in events for t in e.rate.change_times))
        next_change_time = change_times.pop(0) if change_times else np.inf
        if max_time == np.inf:
            assert max_tips is not None
            target_n_tips = rng.integers(min_tips, max_tips + 1)
        else:
            target_n_tips = None

        while current_time < max_time:
            if time.perf_counter() - start_clock > timeout:
                raise TimeoutError("Simulation timed out.")

            events = model.events
            rates = [e.get_propensity(model, current_time) for e in events]

            instantaneous_events = [e for e, r in zip(events, rates) if r == np.inf]
            if instantaneous_events:
                event = instantaneous_events[rng.integers(len(instantaneous_events))]
                event.apply(model, current_time, rng)
                continue

            if (
                not any(rates)
                or max_tips is not None
                and model.n_sampled >= max_tips
                or target_n_tips is not None
                and model.n_sampled >= target_n_tips
            ):
                break

            time_step = rng.exponential(1 / sum(rates))
            if current_time + time_step >= next_change_time:
                current_time = next_change_time
                next_change_time = change_times.pop(0) if change_times else np.inf
                continue
            if current_time + time_step >= max_time:
                current_time = max_time
                break
            current_time += time_step

            event_idx = np.searchsorted(np.cumsum(rates) / sum(rates), rng.random())
            events[int(event_idx)].apply(model, current_time, rng)

        for individual in model.get_population():
            if rng.random() < sampling_probability_at_present:
                model.sample(individual, current_time, True)

        if min_tips <= model.n_sampled and (
            max_tips is None or model.n_sampled <= max_tips
        ):
            return model.get_sampled_tree()


def generate_trees(
    output_dir: str,
    n_trees: int,
    events: Sequence[Event],
    min_tips: int = 1,
    max_tips: int | None = None,
    max_time: float = np.inf,
    init_state: str | None = None,
    sampling_probability_at_present: float = 0.0,
    node_features: Iterable[Feature] | None = None,
    seed: int | None = None,
    n_jobs: int = -1,
    timeout: float = np.inf,
) -> None:
    def _simulate_tree(seed: int) -> Tree:
        while True:
            try:
                tree = simulate_tree(
                    events=events,
                    min_tips=min_tips,
                    max_tips=max_tips,
                    max_time=max_time,
                    init_state=init_state,
                    sampling_probability_at_present=sampling_probability_at_present,
                    seed=seed,
                    timeout=timeout,
                )
                if node_features is not None:
                    set_features(tree, node_features)
                return tree
            except TimeoutError:
                print("Simulation timed out, retrying with a different seed...")
            seed += 1

    if os.path.exists(output_dir):
        raise FileExistsError(f"Output directory {output_dir} already exists")
    os.makedirs(output_dir)

    rng = default_rng(seed)
    jobs = joblib.Parallel(n_jobs=n_jobs, return_as="generator_unordered")(
        joblib.delayed(_simulate_tree)(seed=int(rng.integers(2**32)))
        for _ in range(n_trees)
    )
    for i, tree in tqdm(
        enumerate(jobs), total=n_trees, desc=f"Generating trees in {output_dir}..."
    ):
        dump_newick(tree, os.path.join(output_dir, f"{i}.nwk"))



from phylogenie.treesimulator.events import (
    Birth,
    BirthWithContactTracing,
    Death,
    Event,
    Migration,
    Mutation,
    MutationTargetType,
    Sampling,
    SamplingWithContactTracing,
    get_BD_events,
    get_BDEI_events,
    get_BDSS_events,
    get_canonical_events,
    get_contact_tracing_events,
    get_epidemiological_events,
    get_FBD_events,
    get_mutation_id,
)
from phylogenie.treesimulator.features import Feature, set_features
from phylogenie.treesimulator.gillespie import generate_trees, simulate_tree
from phylogenie.treesimulator.model import get_node_state

__all__ = [
    "Birth",
    "BirthWithContactTracing",
    "Death",
    "Event",
    "Migration",
    "Mutation",
    "MutationTargetType",
    "Sampling",
    "SamplingWithContactTracing",
    "get_BD_events",
    "get_BDEI_events",
    "get_BDSS_events",
    "get_canonical_events",
    "get_contact_tracing_events",
    "get_epidemiological_events",
    "get_FBD_events",
    "generate_trees",
    "simulate_tree",
    "get_mutation_id",
    "get_node_state",
    "Feature",
    "set_features",
]



from typing import Any

from pydantic import BaseModel, ConfigDict


class StrictBaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


class Distribution(BaseModel):
    type: str
    model_config = ConfigDict(extra="allow")

    @property
    def args(self) -> dict[str, Any]:
        assert self.model_extra is not None
        return self.model_extra



from phylogenie.tree import Tree


def get_node_leaf_counts(tree: Tree) -> dict[Tree, int]:
    n_leaves: dict[Tree, int] = {}
    for node in tree.postorder_traversal():
        n_leaves[node] = sum(n_leaves[child] for child in node.children) or 1
    return n_leaves


def get_node_depth_levels(tree: Tree) -> dict[Tree, int]:
    depth_levels: dict[Tree, int] = {}
    for node in tree:
        if node.parent is None:
            depth_levels[node] = 0
        else:
            depth_levels[node] = depth_levels[node.parent] + 1
    return depth_levels


def get_node_depths(tree: Tree) -> dict[Tree, float]:
    depths: dict[Tree, float] = {}
    for node in tree:
        if node.parent is None:
            depths[node] = 0
        else:
            if node.branch_length is None:
                raise ValueError(f"Branch length of node {node.name} is not set.")
            depths[node] = depths[node.parent] + node.branch_length
    return depths


def get_node_height_levels(tree: Tree) -> dict[Tree, int]:
    height_levels: dict[Tree, int] = {}
    for node in tree.postorder_traversal():
        if node.is_leaf():
            height_levels[node] = 0
        else:
            height_levels[node] = max(
                1 + height_levels[child] for child in node.children
            )
    return height_levels


def get_node_heights(tree: Tree) -> dict[Tree, float]:
    heights: dict[Tree, float] = {}
    for node in tree.postorder_traversal():
        if node.is_leaf():
            heights[node] = 0
        else:
            if any(child.branch_length is None for child in node.children):
                raise ValueError(
                    f"Branch length of one or more children of node {node.name} is not set."
                )
            heights[node] = max(
                child.branch_length + heights[child]  # pyright: ignore
                for child in node.children
            )
    return heights



from collections import deque
from collections.abc import Iterator
from typing import Any


class Tree:
    def __init__(self, name: str = "", branch_length: float | None = None):
        self.name = name
        self.branch_length = branch_length
        self._parent: Tree | None = None
        self._children: list[Tree] = []
        self._features: dict[str, Any] = {}

    @property
    def children(self) -> tuple["Tree", ...]:
        return tuple(self._children)

    @property
    def parent(self) -> "Tree | None":
        return self._parent

    @property
    def features(self) -> dict[str, Any]:
        return self._features.copy()

    @property
    def depth(self) -> float:
        if self.parent is None:
            return 0.0
        if self.branch_length is None:
            raise ValueError(f"Branch length of node {self.name} is not set.")
        return self.parent.depth + self.branch_length

    @property
    def depth_level(self) -> int:
        if self.parent is None:
            return 0
        return self.parent.depth_level + 1

    @property
    def height(self) -> float:
        if self.is_leaf():
            return 0.0
        if any(child.branch_length is None for child in self.children):
            raise ValueError(
                f"Branch length of one or more children of node {self.name} is not set."
            )
        return max(
            child.branch_length + child.height  # pyright: ignore
            for child in self.children
        )

    @property
    def height_level(self) -> int:
        if self.is_leaf():
            return 0
        return 1 + max(child.height_level for child in self.children)

    def set(self, key: str, value: Any) -> None:
        self._features[key] = value

    def update_features(self, features: dict[str, Any]) -> None:
        self._features.update(features)

    def get(self, key: str) -> Any:
        return self._features[key]

    def delete(self, key: str) -> None:
        del self._features[key]

    def add_child(self, child: "Tree") -> "Tree":
        child._parent = self
        self._children.append(child)
        return self

    def remove_child(self, child: "Tree") -> None:
        self._children.remove(child)
        child._parent = None

    def set_parent(self, node: "Tree | None"):
        self._parent = node
        if node is not None:
            node._children.append(self)

    def inorder_traversal(self) -> Iterator["Tree"]:
        if self.is_leaf():
            yield self
            return
        if len(self.children) != 2:
            raise ValueError("Inorder traversal is only defined for binary trees.")
        left, right = self.children
        yield from left.inorder_traversal()
        yield self
        yield from right.inorder_traversal()

    def preorder_traversal(self) -> Iterator["Tree"]:
        yield self
        for child in self.children:
            yield from child.preorder_traversal()

    def postorder_traversal(self) -> Iterator["Tree"]:
        for child in self.children:
            yield from child.postorder_traversal()
        yield self

    def breadth_first_traversal(self) -> Iterator["Tree"]:
        queue: deque["Tree"] = deque([self])
        while queue:
            node = queue.popleft()
            yield node
            queue.extend(node.children)

    def get_node(self, name: str) -> "Tree":
        for node in self:
            if node.name == name:
                return node
        raise ValueError(f"Node with name {name} not found.")

    def is_leaf(self) -> bool:
        return not self.children

    def get_leaves(self) -> tuple["Tree", ...]:
        return tuple(node for node in self if node.is_leaf())

    def ladderize(self, feature: str) -> None:
        self._children.sort(key=lambda x: x.get(feature))
        for child in self.children:
            child.ladderize(feature)

    def copy(self):
        new_tree = Tree(self.name, self.branch_length)
        new_tree.update_features(self._features)
        for child in self.children:
            new_tree.add_child(child.copy())
        return new_tree

    def __iter__(self) -> Iterator["Tree"]:
        return self.preorder_traversal()

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __repr__(self) -> str:
        return f"TreeNode(name='{self.name}', branch_length={self.branch_length}, features={self.features})"



from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any

import numpy as np
from numpy.random import Generator

from phylogenie.skyline import SkylineParameterLike, skyline_parameter
from phylogenie.tree import Tree


@dataclass
class Individual:
    id: int
    node: Tree
    state: str


class Event(ABC):
    def __init__(self, state: str, rate: SkylineParameterLike):
        self.state = state
        self.rate = skyline_parameter(rate)

    def draw_individual(self, model: "Model", rng: Generator) -> int:
        return rng.choice(model.get_population(self.state))

    def get_propensity(self, model: "Model", time: float) -> float:
        n_individuals = model.count_individuals(self.state)
        rate = self.rate.get_value_at_time(time)
        if rate == np.inf and not n_individuals:
            return 0
        return rate * n_individuals

    @abstractmethod
    def apply(self, model: "Model", time: float, rng: Generator) -> None: ...


def _get_node_name(node_id: int, state: str) -> str:
    return f"{node_id}|{state}"


def get_node_state(node_name: str) -> str:
    try:
        return node_name.split("|")[1]
    except IndexError:
        raise ValueError(
            f"Invalid node name: {node_name} (expected format 'id|state')."
        )


class Model:
    def __init__(self, init_state: str, events: Sequence[Event]):
        self._next_node_id = 0
        self._next_individual_id = 0
        self._population: dict[int, Individual] = {}
        self._states: dict[str, set[int]] = defaultdict(set)
        self._sampled: set[str] = set()
        self._tree = self._get_new_individual(init_state).node
        self._events = list(events)
        self.context: dict[str, Any] = {}

    @property
    def n_sampled(self) -> int:
        return len(self._sampled)

    @property
    def events(self) -> tuple[Event, ...]:
        return tuple(self._events)

    def add_event(self, event: Event) -> None:
        self._events.append(event)

    def _get_new_node(self, state: str) -> Tree:
        self._next_node_id += 1
        node = Tree(_get_node_name(self._next_node_id, state))
        return node

    def _get_new_individual(self, state: str) -> Individual:
        self._next_individual_id += 1
        individual = Individual(
            self._next_individual_id, self._get_new_node(state), state
        )
        self._population[individual.id] = individual
        self._states[state].add(individual.id)
        return individual

    def _set_branch_length(self, node: Tree, time: float) -> None:
        if node.branch_length is not None:
            raise ValueError(f"Branch length of node {node.name} is already set.")
        node.branch_length = time if node.parent is None else time - node.parent.depth

    def _stem(self, individual: Individual, time: float) -> None:
        self._set_branch_length(individual.node, time)
        stem_node = self._get_new_node(individual.state)
        individual.node.add_child(stem_node)
        individual.node = stem_node

    def remove(self, id: int, time: float) -> None:
        individual = self._population[id]
        self._set_branch_length(individual.node, time)
        self._population.pop(id)
        self._states[individual.state].remove(id)

    def migrate(self, id: int, state: str, time: float) -> None:
        individual = self._population[id]
        self._states[individual.state].remove(id)
        individual.state = state
        self._states[state].add(id)
        self._stem(individual, time)

    def birth_from(self, id: int, state: str, time: float) -> int:
        individual = self._population[id]
        new_individual = self._get_new_individual(state)
        individual.node.add_child(new_individual.node)
        self._stem(individual, time)
        return new_individual.id

    def sample(self, id: int, time: float, removal: bool) -> None:
        individual = self._population[id]
        if removal:
            self._sampled.add(individual.node.name)
            self.remove(id, time)
        else:
            sample_node = self._get_new_node(individual.state)
            sample_node.branch_length = 0.0
            self._sampled.add(sample_node.name)
            individual.node.add_child(sample_node)
            self._stem(individual, time)

    def get_state(self, id: int) -> str:
        return self._population[id].state

    def get_sampled_tree(self) -> Tree:
        tree = self._tree.copy()
        for node in list(tree.postorder_traversal()):
            if node.name not in self._sampled and not node.children:
                if node.parent is None:
                    raise ValueError("No samples in the tree.")
                else:
                    node.parent.remove_child(node)
            elif len(node.children) == 1:
                (child,) = node.children
                child.set_parent(node.parent)
                assert child.branch_length is not None
                assert node.branch_length is not None
                child.branch_length += node.branch_length
                if node.parent is None:
                    return child
                else:
                    node.parent.remove_child(node)
        return tree

    def get_full_tree(self) -> Tree:
        return self._tree.copy()

    def get_population(self, states: str | None = None) -> list[int]:
        if states is None:
            return list(self._population)
        return list(self._states[states])

    def count_individuals(self, states: str | None = None) -> int:
        if states is None:
            return len(self._population)
        return len(self._states[states])



from enum import Enum

import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from matplotlib.axes import Axes

from phylogenie.tree import Tree
from phylogenie.utils import get_node_depths


class Coloring(str, Enum):
    DISCRETE = "discrete"
    CONTINUOUS = "continuous"


def plot_tree(
    tree: Tree,
    ax: Axes | None = None,
    color_by: str | None = None,
    default_color: str = "black",
    coloring: str | Coloring | None = None,
    cmap: str | None = None,
    show_legend: bool = True,
    x_feature: str | None = None,
) -> Axes:
    if ax is None:
        ax = plt.gca()

    xs = (
        get_node_depths(tree)
        if x_feature is None
        else {node: node.get(x_feature) for node in tree}
    )
    ys = {node: i for i, node in enumerate(tree.inorder_traversal())}

    if color_by is not None:
        features = set(node.get(color_by) for node in tree if color_by in node.features)
        if coloring is None and any(isinstance(f, float) for f in features):
            coloring = Coloring.CONTINUOUS
        elif coloring is None:
            coloring = Coloring.DISCRETE

        if coloring == Coloring.DISCRETE:
            if any(isinstance(f, float) for f in features):
                raise ValueError(
                    "Discrete coloring selected but feature values are not all categorical."
                )
            colormap = plt.get_cmap("tab20" if cmap is None else cmap)
            feature_colors = {
                f: mcolors.to_hex(colormap(i)) for i, f in enumerate(features)
            }
            colors = {
                node: (
                    feature_colors[node.get(color_by)]
                    if color_by in node.features
                    else default_color
                )
                for node in tree
            }

            if show_legend:
                legend_handles = [
                    mpatches.Patch(color=feature_colors[f], label=str(f))
                    for f in features
                ]
                if any(color_by not in node.features for node in tree):
                    legend_handles.append(
                        mpatches.Patch(color=default_color, label="None")
                    )
                ax.legend(handles=legend_handles, title=color_by)  # pyright: ignore

        elif coloring == Coloring.CONTINUOUS:
            norm = mcolors.Normalize(vmin=min(features), vmax=max(features))
            colormap = plt.get_cmap("viridis" if cmap is None else cmap)
            colors = {
                node: (
                    colormap(norm(float(node.get(color_by))))
                    if color_by in node.features
                    else default_color
                )
                for node in tree
            }

            sm = plt.cm.ScalarMappable(cmap=colormap, norm=norm)
            ax.get_figure().colorbar(sm, ax=ax)  # pyright: ignore

        else:
            raise ValueError(
                f"Unknown coloring method: {coloring}. Choices are {list(Coloring)}."
            )
    else:
        colors = {node: default_color for node in tree}

    for node in tree:
        x1, y1 = xs[node], ys[node]
        if node.parent is None:
            ax.hlines(y=y1, xmin=0, xmax=x1, color=colors[node])  # pyright: ignore
            continue
        x0, y0 = xs[node.parent], ys[node.parent]
        ax.vlines(x=x0, ymin=y0, ymax=y1, color=colors[node])  # pyright: ignore
        ax.hlines(y=y1, xmin=x0, xmax=x1, color=colors[node])  # pyright: ignore

    ax.set_yticks([])  # pyright: ignore
    return ax
